/**
 * $Id: ClasspathMonitor.java,v 1.30 2009/04/21 19:13:56 oboehm Exp $
 *
 * Copyright (c) 2008 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * (c)reated 10.02.2009 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.monitor;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.jar.JarFile;
import java.util.zip.*;

import javax.management.JMException;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.impl.LogFactoryImpl;

import patterntesting.runtime.annotation.ProfileMe;
import patterntesting.runtime.jmx.MBeanHelper;
import patterntesting.runtime.util.*;

/**
 * To avoid classpath problems like double entries of the same class or resource
 * in the classpath there are several methods available.
 * <br/>
 * To get the boot classpath the sytem property "sun.boot.class.path" is
 * used to get them. This will work of course only for the SunVM.
 *
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 10.02.2009
 * @version $Revision: 1.30 $
 * @see http://www.javaworld.com/javaworld/javatips/jw-javatip105.html
 */
public class ClasspathMonitor extends Thread implements ClasspathMonitorMBean {

    private static final long serialVersionUID = 20090328L;
    private static final Log log = LogFactoryImpl.getLog(ClasspathMonitor.class);
    private final ClassLoader cloader;
    /** This will work of course only for the SunVM */
    private final String[] bootClassPath = getClasspath("sun.boot.class.path");
    private final String[] classpath;
    private String[] loadedClasses = new String[0];
    private final Set<String> unusedClasses;
    private final String[] classpathClasses;
    private List<Class<?>> doubletList = new ArrayList<Class<?>>();
    private static ClasspathMonitor instance;
    
    static {
        instance = new ClasspathMonitor();
    }

    /**
     * @deprecated use {@link #getInstance()}
     */
    public ClasspathMonitor() {
        this.cloader = checkClassloader(this.getClass().getClassLoader());
        this.classpath = getRealClasspath();
        this.unusedClasses = getClasspathClassSet();
        this.classpathClasses = createClasspathClasses();
    }
    
    /**
     * At the beginning unusedClasses contains all classes found in the
     * classpath. So we use this attribute to get the classpathClasses.
     */
    @ProfileMe
    private String[] createClasspathClasses() {
        Set<String> classSet = unusedClasses;
        String[] classpathClasses = new String[classSet.size()];
        classSet.toArray(classpathClasses);
        return classpathClasses;
    }
    
    /**
     * Yes, it is a Singleton because it offers only some services.
     * So we don't need the object twice.
     * 
     * @return the only instance
     */
    public static ClasspathMonitor getInstance() {
        return instance;
    }
    
    public static void registerAsMBean() {
        try {
            MBeanHelper.registerMBean(getInstance());
        } catch (JMException e) {
            if (log.isInfoEnabled()) {
                log.info(instance + " can't be registered as MBean (" + e + ")");
            }
        }
    }
    
    /**
     * @return the classpath as String array
     */
    private String[] getClasspath(String key) {
        String classpath = System.getProperty(key);
        if (classpath == null) {
            log.info(key + " is not set (not a SunVM?)");
            return new String[0];
        }
        return splitClasspath(classpath);
    }
    
    /**
     * We can use the system property "java.class.path" to get the classpath.
     * But this works not inside an application server or serlet enginle
     * (e.g. inside Tomcat) because they have their own classloader to load
     * the classes.
     * <br/>
     * In the past we tried to use the private (and undocoumented) attribute
     * "domains" of the classloader. This works for a normal application but
     * Tomcat's WebappClassLoader listed also classes in the "domains"-Set.
     * Now we will try to detect the different classloader to access some
     * private and secret attributes of this classloader.
     * <br/>
     * At the moment only org.apache.catalina.loader.WebappClassLoader is
     * supported. For all other classloaders the standard approach using the
     * system property "java.class.path" is used.
     * 
     * @return the classpath as String array
     */
    private String[] getRealClasspath() {
        String classloaderName = cloader.getClass().getName();
        if (classloaderName
                .equals("org.apache.catalina.loader.WebappClassLoader")) {
            return getTomcatClasspath();
        } else {
            log.trace("using 'java.class.path' to get classpath...");
            return getClasspath("java.class.path");
        }
    }
    
    private String[] getTomcatClasspath() {
        try {
            Field field = ReflectionHelper.getField(cloader.getClass(),
                    "repositoryURLs");
            URL[] repositoryURLs = (URL[]) field.get(cloader);
            String[] cp = new String[repositoryURLs.length];
            for (int i = 0; i < cp.length; i++) {
                File f = Converter.toFile(repositoryURLs[i].toURI());
                cp[i] = f.getAbsolutePath();
            }
            return cp;
        } catch (Exception e) {
            log.warn("can't access field 'repositoryURLs'", e);
            return getClasspath("java.class.path");
        }
    }

    private ClassLoader checkClassloader(ClassLoader classloader) {
        if (classloader == null) {
            classloader = Thread.currentThread().getContextClassLoader();
            log.warn("null ClassLoader given - replaced by " + classloader);
        }
        return classloader;
    }

    /**
     * @see ClasspathMonitorMBean#whichResource(String)
     */
    @ProfileMe
    public URI whichResource(String name) {
        URL classUrl = ClasspathMonitor.class.getResource(name);
        if (classUrl == null) {
            if (!name.startsWith("/")) {
                return whichResource("/" + name);
            }
            if (log.isTraceEnabled()) {
            	log.trace(name + " not found in classpath");
            }
            return null;
        } else {
            if (log.isTraceEnabled()) {
                log.trace("using " + classUrl + " for " + name);
            }
        }
        try {
        	return classUrl.toURI();
        } catch (URISyntaxException canthappen) {
            String msg = "can't convert " + classUrl + " to URI";
            log.error(msg, canthappen);
            throw new RuntimeException(canthappen);
        }
    }

    /**
     * Returns the URI of the given resource and the given classloader.
     * 
     * @param name
     *            resource name (e.g. "log4j.properties")
     * @param classloader
     * @return URI of the given resource (or null if resource was not found)
     */
    @ProfileMe
    public URI whichResource(String name, ClassLoader classloader) {
        assert classloader != null : "no classloader given";
        URL url = classloader.getResource(name);
        if (url == null) {
            if (name.startsWith("/")) {
                return whichResource(name.substring(1));
            }
            return null;
        }
        try {
            return new URI(url.toString());
        } catch (URISyntaxException canthappen) {
            String msg = "can't convert " + url + " to URI";
            log.error(msg, canthappen);
            throw new RuntimeException(canthappen);
        }
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#whichClass(java.lang.String)
     */
    @ProfileMe
    public URI whichClass(String name) {
        String resource = Converter.classToResource(name);
        return whichResource(resource);
    }
    
    /**
     * Where is the given class found in the classpath? Which class was loaded
     * from which URI?
     * 
     * @param clazz
     *            the class
     * @return file or jar path
     */
    @ProfileMe
    public URI whichClass(Class<?> clazz) {
        return whichClass(clazz.getName());
    }
    
    /**
     * Returns the jar file or path where the given classname was found.
     * 
     * @param classname e.g. java.lang.String
     * @return jar or path as URI
     */
    public URI whichClassPath(String classname) {
        String resource = Converter.classToResource(classname);
        return whichResourcePath(resource);
    }
    
    /**
     * Returns the jar file or path where the given class was found.
     *  
     * @param clazz e.g. Sting.class
     * @return jar or path as URI
     */
    public URI whichClassPath(Class<?> clazz) {
        return whichClassPath(clazz.getName());
    }
    
    /**
     * Returns the jar file or path where the given resource was found
     * 
     * @param resource e.g. log4j.properties
     * @return jar or path as URI
     */
    public URI whichResourcePath(String resource) {
        URI uri = this.whichResource(resource);
        if (uri == null) {
            return null;
        }
        return getParent(uri, resource);
    }
    
    public JarFile whichClassJar(Class<?> clazz) {
        return whichClassJar(clazz.getName());
    }
    
    public JarFile whichClassJar(String classname) {
        String resource = Converter.classToResource(classname);
        return whichResourceJar(resource);
    }
    
    public JarFile whichResourceJar(String resource) {
        return whichResourceJar(this.whichResourcePath(resource));
    }
    
    public static JarFile whichResourceJar(URI resource) {
        if (resource == null) {
            return null;
        }
        File file = Converter.toFile(resource);
        try {
            return new JarFile(file);
        } catch (IOException ioe) {
            log.debug("can't read " + file, ioe);
            return null;
        }        
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * patterntesting.runtime.ClasspathCheckerMBean#getResources(java.lang.String
     * )
     */
    @ProfileMe
    public Enumeration<URL> getResources(String name) {
        try {
            Enumeration<URL> resources = cloader.getResources(name);
            if (!resources.hasMoreElements()) {
                if (name.startsWith("/")) {
                    return getResources(name.substring(1));
                }
                if (log.isDebugEnabled()) {
                    log.debug(name + " not found in classpath");
                }
            }
            return resources;
        } catch (IOException ioe) {
            log.info(name + " not found in classpath", ioe);
            return null;
        }
    }

    public static URL getResource(String name) {
        return ClasspathMonitor.class.getResource(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * patterntesting.runtime.ClasspathCheckerMBean#getNoResources(java.lang
     * .String)
     */
    @ProfileMe
    public int getNoResources(String name) {
        Enumeration<URL> resources = getResources(name);
        if (resources == null) {
            return 0;
        }
        int n = 0;
        while (resources.hasMoreElements()) {
            n++;
            resources.nextElement();
        }
        if (log.isTraceEnabled()) {
        	log.trace(n + " element(s) of " + name + " found in classpath");
        }
        return n;
    }

    public int getNoClasses(Class<?> cl) {
        return getNoClasses(cl.getName());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * patterntesting.runtime.ClasspathCheckerMBean#getNoClasses(java.lang.String
     * )
     */
    public int getNoClasses(String classname) {
        return getNoResources(Converter.classToResource(classname));
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#isDoublet(java.lang.String)
     */
    @ProfileMe
    public boolean isDoublet(String name) {
        Enumeration<URL> resources = getResources(name);
        if ((resources == null) || !resources.hasMoreElements()) {
            throw new NoSuchElementException(name);
        }
        resources.nextElement();
        if (resources.hasMoreElements()) {
            logDoublets(name);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Is the given class a doublet, i.e. can it be found more than once in the
     * classpath?
     * 
     * @param clazz
     *            the class to be checked
     * @return true if class is found more than once in the classpath
     * @throws NoSuchElementException
     *             if class is not found in the classpath
     */
    @ProfileMe
    public boolean isDoublet(Class<?> clazz) {
        String classname = clazz.getName();
        String resource = Converter.classToResource(classname);
        return isDoublet(resource);
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#getFirstDoublet(java.lang
     *      .String)
     */
    public URL getFirstDoublet(String name) {
        return getDoublet(name, 1);
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#getDoublet(java.lang.String,
     *      int)
     */
    public URL getDoublet(String name, int nr) {
        Enumeration<URL> resources = getResources(name);
        for (int i = 0; resources.hasMoreElements(); i++) {
            URL url = (URL) resources.nextElement();
            if (i == nr) {
                return url;
            }
        }
        return null;
    }

    public URL getDoublet(Class<?> clazz, int nr) {
        String resource = Converter.classToResource(clazz.getName());
        return getDoublet(resource, nr);
    }

    private void logDoublets(String name) {
        if (log.isTraceEnabled()) {
            List<URL> doublets = new Vector<URL>();
            Enumeration<URL> resources = getResources(name);
            while (resources.hasMoreElements()) {
                URL url = resources.nextElement();
                doublets.add(url);
            }
            log.trace(name + " doublets found: " + doublets);
        }
    }

    /**
     * Looks for each loaded class if it is a doublet or not.
     * For the performance reason it looks in the doubletList from the
     * last time if it is already found. This is done because normally
     * the number of doublets does not decrease.
     * 
     * @return a sorted list of found doublets
     */
    @ProfileMe
    public synchronized List<Class<?>> getDoubletList() {
        List<Class<?>> loadedClassList = this.getLoadedClassList();
        for (Class<?> clazz : loadedClassList) {
            if (doubletList.contains(clazz)) {
                continue;
            }
            try {
                if (this.isDoublet(clazz)) {
                    doubletList.add(clazz);
                }
            } catch (NoSuchElementException nsee) {
                if (log.isTraceEnabled()) {
                    log.trace(clazz + " not found -> ignored");
                }
            }
        }
        Collections.sort(doubletList, new ObjectComparator());
        return doubletList;
    }
    
    /**
     * @see ClasspathMonitorMBean#getDoublets()
     */
    @ProfileMe
    public String[] getDoublets() {
        log.debug("calculating doublets...");
        List<Class<?>> doubletList = this.getDoubletList();
        String[] doublets = new String[doubletList.size()];
        for (int i = 0; i < doublets.length; i++) {
            doublets[i] = doubletList.get(i).toString();
        }
        return doublets;
    }
    
    /**
     * @see ClasspathMonitorMBean#getDoubletClasspath()
     */
    @ProfileMe
    public String[] getDoubletClasspath() {
        log.debug("calculating doublet-classpath...");
    	Set<URI> classpathSet = getDoubletClasspathSet();
        return getAsArray(classpathSet);
    }

    @ProfileMe
    public SortedSet<URI> getDoubletClasspathSet() {
    	List<Class<?>> doubletList = this.getDoubletList();
    	SortedSet<URI> classpath = new TreeSet<URI>();
    	for (Class<?> cl : doubletList) {
			String resource = Converter.classToResource(cl);
			Enumeration<URL> resources = this.getResources(resource);
			while(resources.hasMoreElements()) {
				URL url = resources.nextElement();
				try {
					URI path = url.toURI();
					classpath.add(getParent(path, resource));
				} catch (URISyntaxException canthappen) {
					log.debug("can't convert " + url + " to URI", canthappen);
				}
			}
		}
    	return classpath;
    }

    /**
     * Dumps the internal fields of the classloader (after an idea from
     * "Java ist auch eine Insel").
     * 
     * @see http 
     *      ://www.galileocomputing.de/openbook/javainsel6/javainsel_21_003.htm
     *      #mj7e2d89f4b96fde1900bc09fd1db83fb1
     */
    @ProfileMe
    public String getClassLoaderDetails() {
        StringBuffer sbuf = new StringBuffer("dump of " + cloader + ":\n");
        for (Class<?> cl = this.cloader.getClass(); cl != null; cl = cl
                .getSuperclass()) {
            sbuf.append('\t');
            dumpFields(sbuf, cl, this.cloader);
        }
        return sbuf.toString().trim();
    }

    @ProfileMe
    private static final void dumpFields(StringBuffer sbuf, Class<?> cl,
            Object obj) {
        Field[] fields = cl.getDeclaredFields();
        AccessibleObject.setAccessible(fields, true);
        for (int i = 0; i < fields.length; i++) {
            sbuf.append(fields[i]);
            sbuf.append(" = ");
            try {
                sbuf.append(fields[i].get(obj));
            } catch (Exception e) {
                sbuf.append("<" + e + ">");
            }
            sbuf.append('\n');
        }
    }

    /**
     * Returns the packages which were loaded by the classloader. <br/>
     * Unfortunately ClassLoader.getPackages() is protected - ok, let's do the
     * hard way using reflexion.
     * 
     * @return array with the loaded packages
     */
    @ProfileMe
    public Package[] getLoadedPackageArray() {
        Method method = getMethod(cloader.getClass(), "getPackages");
        if (method == null) {
            log.warn(cloader.getClass() + "#getPackages() not found");
            return null;
        }
        try {
            method.setAccessible(true);
            return (Package[]) method.invoke(cloader);
        } catch (Exception e) {
            log.warn("can't access " + method, e);
            return null;
        }
    }

    /**
     * @see ClasspathMonitorMBean#getLoadedPackages()
     */
    public String[] getLoadedPackages() {
        Package[] packages = getLoadedPackageArray();
        String[] strings = new String[packages.length];
        for (int i = 0; i < packages.length; i++) {
            strings[i] = packages[i].toString();
        }
        Arrays.sort(strings);
        return strings;
    }

    /**
     * If you want to dump all packages you can use this method. The output will
     * be sorted.
     * 
     * @return each package in a single line
     */
    public String getLoadedPackagesAsString() {
        String[] packages = getLoadedPackages();
        StringBuffer sbuf = new StringBuffer();
        for (int i = 0; i < packages.length; i++) {
            sbuf.append(packages[i]);
            sbuf.append('\n');
        }
        return sbuf.toString().trim();
    }

    private static Method getMethod(Class<?> cl, String name) {
        try {
            return cl.getDeclaredMethod(name);
        } catch (SecurityException e) {
            log.warn("can't get method " + cl + "#" + name, e);
            return null;
        } catch (NoSuchMethodException e) {
            Class<?> superclass = cl.getSuperclass();
            if (superclass == null) {
                return null;
            } else {
                return getMethod(superclass, name);
            }
        }
    }

    /**
     * Returns a list of classes which were loaded by the given classloader. <br/>
     * Ok, we must do some hacks here: there is an undocumented attribute
     * "classes" which contains the loaded classes. <br/>
     * HANDLE WITH CARE (it's a hack and it depends on the used classloader)
     * 
     * @param cloader
     *            classloader
     * @return list of classes
     */
    @ProfileMe
    @SuppressWarnings("unchecked")
    private static List<Class<?>> getLoadedClassList(ClassLoader cloader) {
        try {
            Field field = ReflectionHelper.getField(cloader.getClass(),
                    "classes");
            List<Class<?>> classList = (List<Class<?>>) field.get(cloader);
            return new ArrayList<Class<?>>(classList);
        } catch (Exception e) {
            log.warn("can't access field 'classes'", e);
            return Collections.EMPTY_LIST;
        }
    }

    /**
     * Returns a list of classes which were loaded by the classloader.
     * 
     * @return list of classes
     */
    @ProfileMe
    synchronized public List<Class<?>> getLoadedClassList() {
        List<Class<?>> classlist = getLoadedClassList(cloader);
        return classlist;
    }

    /**
     * As MBean a string array could be displayed by the 'jconsole'. A class
     * array not.
     * 
     * @return the classes as sorted string array
     */
    synchronized public String[] getLoadedClasses() {
        List<Class<?>> classes = getLoadedClassList();
        if (classes.size() != loadedClasses.length) {
            loadedClasses = new String[classes.size()];
            for (int i = 0; i < loadedClasses.length; i++) {
                loadedClasses[i] = classes.get(i).toString();
            }
            Arrays.sort(loadedClasses);
        }
        return loadedClasses;
    }

    @ProfileMe
    public String getLoadedClassesAsString() {
        List<Class<?>> classes = getLoadedClassList();
        Collections.sort(classes, new ObjectComparator());
        StringBuffer sbuf = new StringBuffer();
        for (Iterator<Class<?>> i = classes.iterator(); i.hasNext();) {
            sbuf.append(i.next().toString().trim());
            sbuf.append('\n');
        }
        return sbuf.toString().trim();
    }
    
    /**
     * Unused classes are classes which are not loaded but which are found
     * in the classpath.
     * 
     * @return unused classes
     * @see #getLoadedClasses()
     * @see #getClasspathClasses()
     */
    @ProfileMe
    public String[] getUnusedClasses() {
        List<Class<?>> classes = getLoadedClassList();
        Collection<String> used = new ArrayList<String>();
        for (Class<?> cl : classes) {
            String classname = cl.getName();
            if (unusedClasses.contains(classname)) {
                used.add(classname);
            }
        }
        unusedClasses.removeAll(used);
        String[] unused = new String[unusedClasses.size()];
        unusedClasses.toArray(unused);
        return unused;
    }
    
    /**
     * Scans the classpath for all classes.
     * 
     * @return all classes as String array
     */
    public String[] getClasspathClasses() {
        return classpathClasses;
    }
    
    @ProfileMe
    private Set<String> getClasspathClassSet() {
        Set<String> classSet = new TreeSet<String>();
        for (int i = 0; i < classpath.length; i++) {
            addClasses(classSet, new File(classpath[i]));
        }
        return classSet;
    }
    
    private static void addClasses(Set<String> classSet, File path) {
        if (log.isTraceEnabled()) {
            log.trace("adding classes from " + path.getAbsolutePath() + "...");
        }
        try {
            if (path.isDirectory()) {
                addClassesFromDir(classSet, path);
            } else {
                addClassesFromArchive(classSet, path);
            }
        } catch (IOException e) {
            log.warn("can't add classes from " + path.getAbsolutePath());
        }
    }
    
    @ProfileMe
    private static void addClassesFromDir(Set<String> classSet, File dir)
            throws IOException {
        ClassWalker classWalker = new ClassWalker(dir);
        Collection<String> classes = classWalker.getClasses();
        classSet.addAll(classes);
    }
    
    @ProfileMe
    private static void addClassesFromArchive(Set<String> classSet, File archive)
            throws IOException {
        ZipFile zipFile = new ZipFile(archive);
        Enumeration<? extends ZipEntry> entries = zipFile.entries();
        while(entries.hasMoreElements()) {
            ZipEntry entry = entries.nextElement();
            String name = entry.getName();
            if (name.endsWith(".class")) {
                classSet.add(Converter.resourceToClass(name));
            }
        }
    }

    /**
     * Gets the loaded classpath (without the bootclasspath) as sorted set.
     * <br/>
     * TODO: this method lets place for more performance 
     *       (e.g. using a Hashtable<Class, URI> for caching)
     * 
     * @return the loaded classpath (excluding the bootclasspath)
     */
    @ProfileMe
    public SortedSet<URI> getUsedClasspathSet() {
        List<Class<?>> loadedClassList = this.getLoadedClassList();
        SortedSet<URI> usedClasspathSet = new TreeSet<URI>();
        for (Class<?> clazz : loadedClassList) {
            URI classUri = this.whichClass(clazz);
            if (classUri != null) {
                URI classpath = getParent(classUri, clazz);
                usedClasspathSet.add(classpath);
            }
        }
        return usedClasspathSet;
    }

    /**
     * It might be that the used classpath changes during the calculation of it.
     * So it is only a good approximation how it looks at the moment of the
     * call.
     * 
     * @see ClasspathMonitor#getUsedClasspath()
     */
    @ProfileMe
    public String[] getUsedClasspath() {
        log.debug("calculating used classpath...");
        SortedSet<URI> classpathSet = this.getUsedClasspathSet();
        return getAsArray(classpathSet);
    }
    
    private static String[] getAsArray(Set<URI> classpathSet) {
        String[] classpath = new String[classpathSet.size()];
        Object[] a = classpathSet.toArray();
        for (int i = 0; i < classpath.length; i++) {
            URI uri = (URI) a[i];
            classpath[i] = uri.getPath();
            if (StringUtils.isEmpty(classpath[i])) {
                classpath[i] = StringUtils.substringAfterLast(uri.toString(),
                        ":");
            }
            if (classpath[i].endsWith(File.separator)) {
                classpath[i] = classpath[i].substring(0,
                        (classpath[i].length() - 1));
            }
        }
        Arrays.sort(classpath);
        return classpath;
    }
    
    /**
     * The unused classpath is this path which are not used in past.
     * <br/>
     * TODO: the first three calls can be parallized (for better performance)
     * 
     * @return the unused classpath (sorted)
     */
    @ProfileMe
    public String[] getUnusedClasspath() {
        log.debug("calculating unused classpath...");
        SortedSet<String> unused = new TreeSet<String>();
        String[] classpath = this.getClasspath();
        log.trace(Arrays.class + " loaded (to get corrected used classpath");
        String[] used = this.getUsedClasspath();
        for (int i = 0; i < classpath.length; i++) {
            String path = classpath[i];
            if (Arrays.binarySearch(used, path) < 0) {
                unused.add(path);
            }
        }
        String[] a = new String[unused.size()];
        return unused.toArray(a);
    }

    
    /**
     * To get the boot classpath the sytem property "sun.boot.class.path" is
     * used to get them. This will work of course only for the SunVM.
     * 
     * @return the boot classpath as String array
     */
    public String[] getBootClasspath() {
        return this.bootClassPath;
    }
    
    /**
     * The classpath is stored in the system property "java.class.path".
     * And this is the classpath which will be returned.
     * 
     * @return the classpath as String array
     */
    public String[] getClasspath() {
        return this.classpath;
    }
    
    private static String[] splitClasspath(String classpath) {
        String separator = System.getProperty("path.separator", ":");
        String[] cp = StringUtils.split(classpath, separator);
        for (int i = 0; i < cp.length; i++) {
            if (cp[i].endsWith(File.separator)) {
                cp[i] = cp[i].substring(0, (cp[i].length() - 1));
            }
        }
        return cp;
    }

    @ProfileMe
    private static URI getParent(URI path, Class<?> clazz) {
        String resource = Converter.classToResource(clazz);
        return getParent(path, resource);
    }

    /**
     * Don't use @ProfileMe here!
     * This method is also used when ProfileStatistic is set up!
     *
     * @param path e.g. "jar:file:/a/b/c.jar!/d/e.class
     * @param resource e.g. "/d/e.class"
     * @return e.g. "jar:file:/a/b/c.jar"
     */
    public static URI getParent(URI path, String resource) {
        String parent = StringUtils.removeEnd(path.toString(), resource);
        if (parent.endsWith(File.separator)) {
            parent = parent.substring(0, (parent.length() - 1));
        }
        if (parent.endsWith("!")) {
            parent = parent.substring(0, (parent.length() - 1));
        }
        try {
            return new URI(parent);
        } catch (URISyntaxException canthappen) {
            log.error("can't extract " + resource + " from " + path,
                            canthappen);
            return path;
        }
    }
    
    /**
     * You can register the instance as shutdown hook. If the VM is
     * terminated the attributes are logged and dumped to a text file in the
     * tmp directory.
     * 
     * @see #logStatistic()
     * @see #dumpStatistic()
     */
    public static void addAsShutdownHook() {
        Runtime.getRuntime().addShutdownHook(instance);
        if (log.isDebugEnabled()) {
            log.debug(instance + " registered as shutdown hook");
        }
    }
    
    /**
     * This method is called when the ClasspathMonitor is registered as
     * shutdown hook.
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public void run() {
        dumpMe();
    }

    /**
     * Logs the different array to the log output.
     */
    public void logMe() {
        logArray("BootClasspath",    getBootClasspath());
        logArray("Classpath",        getClasspath());
        logArray("ClasspathClasses", getClasspathClasses());
        logArray("DoubletClasspath", getDoubletClasspath());
        logArray("Doublets",         getDoublets());
        logArray("LoadedClasses",    getLoadedClasses());
        logArray("LoadedPackages",   getLoadedPackages());
        logArray("UnusedClasses",    getUnusedClasses());
        logArray("UnusedClasspath",  getUnusedClasspath());
        logArray("UsedClasspath",    getUsedClasspath());
    }

    private static void logArray(String title, String[] array) {
        log.info("=== " + title + " ===");
        for (int i = 0; i < array.length; i++) {
            log.info(array[i]);
        }
    }
    
    /**
     * This operation dumps the different MBean attributes to a temporary
     * file with the prefix "cpmon" (for ClasspathMonitor) and the
     * extension ".txt".
     */
    public void dumpMe() {
        try {
            File dumpFile = File.createTempFile("cpmon", ".txt");
            BufferedWriter writer = new BufferedWriter(new FileWriter(dumpFile));
            dumpArray(writer, "BootClasspath",    getBootClasspath());
            dumpArray(writer, "Classpath",        getClasspath());
            dumpArray(writer, "ClasspathClasses", getClasspathClasses());
            dumpArray(writer, "DoubletClasspath", getDoubletClasspath());
            dumpArray(writer, "Doublets",         getDoublets());
            dumpArray(writer, "LoadedClasses",    getLoadedClasses());
            dumpArray(writer, "LoadedPackages",   getLoadedPackages());
            dumpArray(writer, "UnusedClasses",    getUnusedClasses());
            dumpArray(writer, "UnusedClasspath",  getUnusedClasspath());
            dumpArray(writer, "UsedClasspath",    getUsedClasspath());
            writer.close();
            log.info("ClasspathMonitor attributes dumped to " + dumpFile);
        } catch (IOException ioe) {
            log.info("can't dump ClasspathMonitor attributes (" + ioe + ")");
        }
    }
    
    private static void dumpArray(BufferedWriter writer, String title,
            String[] array) throws IOException {
        writer.write("=== ");
        writer.write(title);
        writer.write(" ===");
        writer.newLine();
        for (int i = 0; i < array.length; i++) {
            writer.write(array[i]);
            writer.newLine();
        }
        writer.newLine();
    }
    
    /**
     * We don't want to use the toString() implementation of the Thread class
     * so we use our own one.
     * 
     * @see java.lang.Thread#toString()
     */
    @Override
    public String toString() {
        return this.getClass().getSimpleName() + " for " + this.cloader;
    }

    /**
     * This main method is only for testing the ClasspathMonitor together with
     * the 'jconsole'. On a Java5 VM start it with the following options:
     * -Dcom.sun.management.jmxremote.local.only=false
     * -Dcom.sun.management.jmxremote
     * 
     * @throws JMException
     */
    public static void main(String[] args) throws JMException {
        ClasspathMonitor.registerAsMBean();
        ThreadUtil.sleep(300, TimeUnit.SECONDS);
    }

}

/*
 * $Log: ClasspathMonitor.java,v $
 * Revision 1.30  2009/04/21 19:13:56  oboehm
 * reduced logging for registerAsMBean()
 *
 * Revision 1.29  2009/04/21 19:05:23  oboehm
 * registerAsMBean() is now static
 *
 * Revision 1.28  2009/04/21 18:28:55  oboehm
 * log4j.Logger replaced by commons-logging.Log
 *
 * Revision 1.27  2009/04/16 14:02:03  oboehm
 * new JMX attribute "unusedClasses" added
 *
 * Revision 1.26  2009/04/15 19:25:14  oboehm
 * a new attribute "classpathClasses" added for JMX
 *
 * Revision 1.25  2009/04/14 13:23:45  oboehm
 * getClasspath() works now also inside Tomcat 6.0
 * (and I guess also inside Tomcat 5.5)
 *
 * Revision 1.24  2009/04/13 18:24:32  oboehm
 * ClasspathMonitor is now public again
 * (but marked as @Deprecated)
 *
 * Revision 1.23  2009/04/13 15:41:35  oboehm
 * implementation for getClasspath() now based on domains attribute
 * of ClassLoader
 *
 * Revision 1.22  2009/04/13 14:40:00  oboehm
 * dumpMe operation added
 *
 * Revision 1.21  2009/04/13 13:55:32  oboehm
 * logMe operation added
 *
 * Revision 1.20  2009/04/13 13:32:47  oboehm
 * some small refactorings
 *
 * Revision 1.19  2009/04/11 12:05:31  oboehm
 * some more checks for getClasspath() added
 * (because implementation will change in the future)
 *
 * $Source: /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java/patterntesting/runtime/monitor/ClasspathMonitor.java,v $
 */
