package patterntesting.runtime.monitor;

import java.io.Serializable;
import java.net.*;

public interface ClasspathMonitorMBean extends Serializable {

    /**
     * Looks if the given resource can be found in the classpath.
     * <br/>
     * To avoid problems like "java.rmi.UnmarshalException: failed to unmarshal
     * class java.lang.Object; nested exception is: java.io.IOException: unknown
     * protocol: zip" no longer a URL but URI is now returned
     * 
     * @param name of a resource e.g. "log4j.properties"
     * @return URI of the given resource (or null if resource was not found)
     */
    public URI whichResource(String name);

    /**
     * Looks if the given classname can be found in the classpath.
     * <br/>
     * To avoid problems like "java.rmi.UnmarshalException: failed to unmarshal
     * class java.lang.Object; nested exception is: java.io.IOException: unknown
     * protocol: zip" no longer a URL but URI is now returned
     * 
     * @param name of a class e.g. "java.lang.String"
     * @return URI of the given resource (or null if resource was not found)
     */
    public URI whichClass(String name);

    /**
     * Get number of resources.
     * 
     * @param name of the resource
     * @return how often the resource was found
     */
    public int getNoResources(String name);

    /**
     * Get the number of different versions of the given class.
     * 
     * @param classname
     * @return how often the classname was found in the classpath
     */
    public int getNoClasses(String classname);

    /**
     * Is the given classname or resource a doublet, e.g. can it be found
     * several times in the classpath?
     * If the classname or resource is not in the classpath a
     * java.util.NoSuchElementException will be thrown.
     * 
     * @param name a classname or resource
     * @return true if more than one classname or resource was found in the
     *         classpath
     * @throws NoSuchElementException if no classname or resource was found
     */
    public boolean isDoublet(String name);

    /**
     * Returns the first doublet of the given classname or resource.
     * 
     * @param name a classname or resource
     * @return
     */
    public URL getFirstDoublet(String name);
    
    /**
     * Looks for each loaded class if it is a doublet or not.
     * 
     * @return a sorted array with the found doublets
     */
    public String[] getDoublets();
    
    /**
     * Looks for each found doublet in which classpath it was found.
     * 
     * @return the classpath where doublets were found
     */
    public String[] getDoubletClasspath();

    /**
     * Returns the n'th doublet of the given classname or resource
     * 
     * @param name a classname or resource
     * @param n
     * @return
     */
    public URL getDoublet(String name, int n);

    /**
     * Returns the packages which were loaded by the classloader.
     * The loaded packages are returned as string array so that it can
     * be displayed by the 'jconsole'.
     * 
     * @return the packages as string array
     */
    public String[] getLoadedPackages();
    
    /**
     * Returns the classes which were loaded by the classloader.
     * The loaded packages are returned as string array so that it can
     * be displayed by the 'jconsole'.
     * 
     * @return the classes as string array
     */
    public String[] getLoadedClasses();
    
    /**
     * Returns the classes which are not yet loaded. These are the classes
     * returned by getClasspathClasses() but not by getLoadedClasses().
     * 
     * @return the classes which are not yet loaded
     */
    public String[] getUnusedClasses();
    
    /**
     * Gets the loaded classpath (without the bootclasspath) as URI array.
     * 
     * @return the loaded classpath (sorted)
     */
    public String[] getUsedClasspath();
    
    /**
     * The unused classpath is this path which are not used in past.
     * 
     * @return the unused classpath (sorted)
     */
    public String[] getUnusedClasspath();
    
    /**
     * Gets the boot classpath.
     * 
     * @return boot classpath
     */
    public String[] getBootClasspath();
    
    /**
     * Gets the normal classpath which is also available by the
     * system property "java.class.path".
     * 
     * @return the classpath as String array
     */
    public String[] getClasspath();
    
    /**
     * Looks in the normal classpath after all classes and returns it as
     * String array.
     * 
     * @return all found classes in the classpath
     */
    public String[] getClasspathClasses();
    
    /**
     * Prints the different MBean attributes to the log output.
     */
    public void logMe();
    
    /**
     * This operation dumps the different MBean attributes to a temporary
     * file with the prefix "cpmon" (for ClasspathMonitor) and the
     * extension ".txt".
     */
    public void dumpMe();

}

/*
 * $Log: ClasspathMonitorMBean.java,v $
 * Revision 1.9  2009/04/16 14:02:03  oboehm
 * new JMX attribute "unusedClasses" added
 *
 * Revision 1.8  2009/04/15 19:25:14  oboehm
 * a new attribute "classpathClasses" added for JMX
 *
 * Revision 1.7  2009/04/13 14:40:00  oboehm
 * dumpMe operation added
 *
 * Revision 1.6  2009/04/13 13:55:32  oboehm
 * logMe operation added
 *
 * Revision 1.5  2009/02/21 21:07:45  oboehm
 * getDoubletClasspath() added
 *
 * $Source: /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java/patterntesting/runtime/monitor/ClasspathMonitorMBean.java,v $
 */
