/**
 * $Id: ProfileStatistic.java,v 1.25 2009/04/25 21:13:45 oboehm Exp $
 *
 * Copyright (c) 2008 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * (c)reated 22.12.2008 by oliver (ob@oasd.de)
 */
package patterntesting.runtime.monitor;

import java.io.*;
import java.lang.management.ManagementFactory;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;

import javax.management.*;
import javax.management.openmbean.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.impl.LogFactoryImpl;
import org.aspectj.lang.Signature;

import patterntesting.annotation.check.runtime.MayReturnNull;
import patterntesting.runtime.annotation.DontProfileMe;
import patterntesting.runtime.util.SignatureHelper;

/**
 * This is constructed as a thin layer around com.jamonapi.MonitorFactory for
 * the needs of patterntesting. The reason for this layer is that sometimes you
 * don't want to minimize the use of other libraries. So this implementation
 * provides also an implementation if the JaMon library is missing.
 * 
 * @see com.jamonapi.MonitorFactory
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 22.12.2008
 * @version $Revision: 1.25 $
 */
public class ProfileStatistic extends Thread implements ProfileStatisticMBean {
    
    public static final ObjectName mbeanName;
    private static final ProfileStatistic instance;
    private static final Log log = LogFactoryImpl.getLog(ProfileStatistic.class);
    private static final SimpleProfileMonitor rootMonitor;
    private static final boolean jamonAvailable;
    
    
    
    /////   initialization   //////////////////////////////////////////////////
    
    /**
     * rootMonitor *must* be initialized before isJamon24Available() is called.
     * Otherwise you'll get a NullPointerException because in
     * Jamon24Available() rootMonitor will be accessed (so rootMonitor must
     * be initialized before!)
     */
    static {
        mbeanName = createObjectName();
        rootMonitor = new SimpleProfileMonitor();
        jamonAvailable = isJamon24Available();
        instance = new ProfileStatistic();
    }
    
    /**
     * It is only tested for Jamon 2.4 so we look for it
     * @return true if Jamon 2.4 was found
     */
    protected static boolean isJamon24Available() {
        try {
            String resource = "/com/jamonapi/MonitorFactory.class";
            URL classURL = ProfileStatistic.class.getResource(resource);
            URI jarURI = ClasspathMonitor.getParent(classURL.toURI(), resource);
            JarFile jarfile = ClasspathMonitor.whichResourceJar(jarURI);
            Manifest manifest = jarfile.getManifest();
            Attributes attributes = manifest.getMainAttributes();
            String version = attributes.getValue("version");
            if (version.equalsIgnoreCase("JAMon 2.4")) {
                log.info("JAMon 2.4 available for profiling");
                return true;
            } else {
                log.info(version
                   + " not supported (only JAMon 2.4), using simple profiling");
                return false;
            }
        } catch (Exception e) {
            // if JAMon is not in classpath we come with a NPE to here
            log.info("JAMon not available, using simple profiling");
            return false;
        }
    }

    private static ObjectName createObjectName() {
        String name = "patterntesting.runtime.monitor:type=ProfileStatistic";
        ObjectName objectName = null;
        try {
            objectName = new ObjectName(name);
        } catch (MalformedObjectNameException e) {
            log.info("can't create object name + '" + name + "'");
        }
        return objectName;
    }
    
    protected ProfileStatistic() {
        this.registerAsMBean();
    }
    
    private void registerAsMBean() {
        MBeanServer server = ManagementFactory.getPlatformMBeanServer();
        try {
            server.registerMBean(this, mbeanName);
            log.info(mbeanName + " successful registered as MBean");
        } catch (Exception e) {
            log.info(e);
        }        
    }
    
    public static ProfileStatistic getInstance() {
        return instance;
    }
    
    /**
     * You can register the instance as shutdown hook. If the VM is
     * terminated the profile values are logged and dumped to a CSV file in the
     * tmp directory.
     * 
     * @see #logStatistic()
     * @see #dumpStatistic()
     */
    public static void addAsShutdownHook() {
        Runtime.getRuntime().addShutdownHook(instance);
        if (log.isDebugEnabled()) {
            log.debug(instance + " registered as shutdown hook");
        }
    }
    
    /**
     * We can't reset all ProfileMonitors - we must keep the empty
     * monitors with 0 hits to see which methods or constructors are
     * never called.
     */
    public void reset() {
        List<String> labels = new ArrayList<String>();
        ProfileMonitor[] monitors = getMonitors();
        for (int i = 0; i < monitors.length; i++) {
            if (monitors[i].getHits() == 0) {
                labels.add(monitors[i].getLabel());
            }
        }        
        if (jamonAvailable) {
            JamonMonitorFactory.reset();
            JamonMonitorFactory.addMonitors(labels);
        } else {
            rootMonitor.reset();
            rootMonitor.addChildren(labels);
        }
    }
    
    /**
     * For each constructor and for each method of the given class a
     * ProfileMonitor is initialized. This is done to be able to find
     * constructors and methods which are are never used (i.e. their
     * hit count is zero).
     * 
     * @param sig the given class (faced as signature)
     */
    public static void init(Class<?> cl) {
        if (log.isTraceEnabled()) {
            log.trace("initializing monitors for " + cl + "...");
        }
        init(cl, cl.getMethods());
        init(cl.getConstructors());
    }
    
    /**
     * Only methods of the given class but not the methods of the superclass
     * are initialized for profiling.
     * 
     * @param cl
     * @param methods
     */
    private static void init(Class<?> cl, Method[] methods) {
        for (int i = 0; i < methods.length; i++) {
            Class<?> declaring = methods[i].getDeclaringClass();
            if (cl.equals(declaring)) {
                init(methods[i]);
            } else if (log.isTraceEnabled()) {
                log.trace(methods[i] + " not defined in " + cl
                        + " -> no monitor initialized");
            }
        }
    }
    
    private static void init(Method method) {
        if (method.getAnnotation(DontProfileMe.class) != null) {
            if (log.isTraceEnabled()) {
                log.trace("@DontProfileMe " + method + " is ignored");
            }
            return;
        }
        Signature sig = SignatureHelper.getAsSignature(method);
        ProfileMonitor mon = getMonitor(sig);
        if (log.isTraceEnabled()) {
            log.trace(mon + " initialized");
        }
    }
    
    private static void init(Constructor<?>[] ctors) {
        for (int i = 0; i < ctors.length; i++) {
            init(ctors[i]);
        }
    }
    
    private static void init(Constructor<?> ctor) {
        if (ctor.getAnnotation(DontProfileMe.class) != null) {
            if (log.isTraceEnabled()) {
                log.trace("@DontProfileMe " + ctor + " is ignored");
            }
            return;
        }
        Signature sig = SignatureHelper.getAsSignature(ctor);
        ProfileMonitor mon = getMonitor(sig);
        if (log.isTraceEnabled()) {
            log.trace(mon + " initialized");
        }
    }
    
    
    
    /////   business logic (measurement, statistics and more)   ///////////////
    
    /**
     * This method is called when the PerformanceMonitor is registered as
     * shutdown hook.
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public void run() {
        dumpStatistic();
    }

    public static ProfileMonitor start(Signature sig) {
        ProfileMonitor mon = null;
        if (jamonAvailable) {
            mon = JamonMonitorFactory.getMonitor(sig);
        } else {
            SimpleProfileMonitor parent = getSimpleProfileMonitor(sig);
            mon = new SimpleProfileMonitor(sig, parent);
        }
        mon.start();
        return mon;
    }
    
    private static ProfileMonitor getMonitor(Signature sig) {
        if (jamonAvailable) {
            return JamonMonitorFactory.getMonitor(sig);
        } else {
            return getSimpleProfileMonitor(sig);
        }
    }
    
    private static SimpleProfileMonitor getSimpleProfileMonitor(Signature sig) {
        SimpleProfileMonitor monitor = rootMonitor.getMonitor(sig);
        if (monitor == null) {
            monitor = new SimpleProfileMonitor(sig, rootMonitor);
        }
        return monitor;
    }
    
    private static ProfileMonitor[] getMonitors() {
        if (jamonAvailable) {
            return JamonMonitorFactory.getMonitors();
        } else {
            return rootMonitor.getMonitors();
        }
    }
    
    /**
     * @return monitors sorted after total time (descending order)
     */
    protected static ProfileMonitor[] getSortedMonitors() {
        ProfileMonitor[] monitors = getMonitors();
        Arrays.sort(monitors);
        return monitors;
    }
    
    private ProfileMonitor getMaxHitsMonitor() {
        ProfileMonitor[] monitors = getMonitors();
        ProfileMonitor max = monitors[0];
        for (int i = 1; i < monitors.length; i++) {
            if (monitors[i].getHits() > max.getHits()) {
                max = monitors[i];
            }
        }
        return max;
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileStatisticMBean#getMaxHits()
     */
    public int getMaxHits() {
        return getMaxHitsMonitor().getHits();
    }
    
    public String getMaxHitsLabel() {
        return getMaxHitsMonitor().getLabel();
    }
    
    public String getMaxHitsStatistic() {
        return getMaxHitsMonitor().toShortString();
    }
    
    private ProfileMonitor getMaxTotalMonitor() {
        ProfileMonitor[] monitors = getMonitors();
        ProfileMonitor max = monitors[0];
        for (int i = 1; i < monitors.length; i++) {
            if (monitors[i].getTotal() > max.getTotal()) {
                max = monitors[i];
            }
        }
        return max;
    }

    public double getMaxTotal() {
        return getMaxTotalMonitor().getTotal();
    }
    
    public String getMaxTotalLabel() {
        return getMaxTotalMonitor().getLabel();
    }
    
    public String getMaxTotalStatistic() {
        return getMaxTotalMonitor().toShortString();
    }

    private ProfileMonitor getMaxAvgMonitor() {
        ProfileMonitor[] monitors = getMonitors();
        ProfileMonitor max = monitors[0];
        for (int i = 1; i < monitors.length; i++) {
            if (monitors[i].getAvg() > max.getAvg()) {
                max = monitors[i];
            }
        }
        return max;
    }

    public double getMaxAvg() {
        return getMaxAvgMonitor().getAvg();
    }
    
    public String getMaxAvgLabel() {
        return getMaxAvgMonitor().getLabel();
    }

    public String getMaxAvgStatistic() {
        return getMaxAvgMonitor().toShortString();
    }

    private ProfileMonitor getMaxMaxMonitor() {
        ProfileMonitor[] monitors = getMonitors();
        ProfileMonitor max = monitors[0];
        for (int i = 1; i < monitors.length; i++) {
            if (monitors[i].getMax() > max.getMax()) {
                max = monitors[i];
            }
        }
        return max;
    }

    public double getMaxMax() {
        return getMaxMaxMonitor().getMax();
    }
    
    public String getMaxMaxLabel() {
        return getMaxMaxMonitor().getLabel();
    }
    
    public String getMaxMaxStatistic() {
        return getMaxMaxMonitor().toShortString();
    }

    @SuppressWarnings("unchecked")
    public TabularData getStatistics() {
        try {
            String[] itemNames = { "Label", "Units", "Hits", "Avg", "Total",
                    "Min", "Max", "Active", "AvgActive", "MaxActive",
                    "FirstAccess", "LastAccess" };
            String[] itemDescriptions = { "method name", "time unit (e.g. ms)",
                    "number of hits", "average time", "total time",
                    "minimal time", "maximal time", "active threads",
                    "average number of threads", "maximal number of threads",
                    "first access", "last access" };
            OpenType[] itemTypes = { SimpleType.STRING, SimpleType.STRING,
                    SimpleType.INTEGER, SimpleType.DOUBLE, SimpleType.DOUBLE,
                    SimpleType.DOUBLE, SimpleType.DOUBLE, SimpleType.DOUBLE,
                    SimpleType.DOUBLE, SimpleType.DOUBLE, SimpleType.DATE,
                    SimpleType.DATE };
            CompositeType rowType = new CompositeType("propertyType",
                    "property entry", itemNames, itemDescriptions, itemTypes);
            TabularType tabularType = new TabularType("propertyTabularType",
                    "properties tabular", rowType, itemNames);
            TabularDataSupport data = new TabularDataSupport(tabularType);
            ProfileMonitor[] monitors = getSortedMonitors();
            if (monitors == null) {
                log.warn("can't find monitors");
                return null;
            }
            for (int i = 0; i < monitors.length; i++) {
                Map<String, Object> map = new HashMap<String, Object>();
                map.put("Label", monitors[i].getLabel());
                map.put("Units", monitors[i].getUnits());
                map.put("Hits", monitors[i].getHits());
                map.put("Avg", monitors[i].getAvg());
                map.put("Total", monitors[i].getTotal());
                map.put("Min", monitors[i].getMin());
                map.put("Max", monitors[i].getMax());
                map.put("Active", monitors[i].getActive());
                map.put("AvgActive", monitors[i].getAvgActive());
                map.put("MaxActive", monitors[i].getMaxActive());
                map.put("FirstAccess", monitors[i].getFirstAccess());
                map.put("LastAccess", monitors[i].getLastAccess());
                CompositeDataSupport compData = new CompositeDataSupport(
                        rowType, map);
                data.put(compData);
            }
            return data;
        } catch (OpenDataException e) {
            log.error("can't create TabularData for log settings", e);
            return null;
        }
    }

    public void logStatistic() {
        log.info("----- Profile Statistic -----");
        ProfileMonitor[] monitors = getSortedMonitors();
        for (ProfileMonitor profMon : monitors) {
            log.info(profMon);
        }
    }
    
    public void dumpStatistic() {
        try {
            File dumpFile = File.createTempFile("profile", ".csv");
            BufferedWriter writer = new BufferedWriter(new FileWriter(dumpFile));
            ProfileMonitor[] monitors = getSortedMonitors();
            writer.write(monitors[0].toCsvHeadline());
            writer.newLine();
            for (ProfileMonitor profMon : monitors) {
                writer.write(profMon.toCsvString());
                writer.newLine();
            }
            writer.close();
            log.info("profiling data dumped to " + dumpFile);
        } catch (IOException ioe) {
            log.info("can't dump statistic (" + ioe + ")");
        }
    }

    /**
     * Do you want to look for the monitor of a given method? Use this
     * method here.
     * 
     * @param clazz
     * @param method the method name, including parameter
     *        e.g. "getProfileMonitor(Class,String)"
     * @return
     */
    @MayReturnNull
    public ProfileMonitor getProfileMonitor(Class<?> clazz, String method) {
        return this.getProfileMonitor(clazz.getName() + "." + method);
    }
    
    /**
     * Do you want to look for the monitor of a given method? Use this
     * method here.
     * 
     * @param signature e.g. "hello.World(String[])"
     * @return
     */
    @MayReturnNull
    public ProfileMonitor getProfileMonitor(String signature) {
        for (ProfileMonitor profMon : getMonitors()) {
            if (signature.equals(profMon.getLabel())) {
                return profMon;
            }
        }
        log.trace("no ProfileMonitor for " + signature + " found");
        return null;
    }

    /* (non-Javadoc)
     * @see java.lang.Thread#toString()
     */
    @Override
    public String toString() {
        return mbeanName.toString();
    }

}

/**
 * $Log: ProfileStatistic.java,v $
 * Revision 1.25  2009/04/25 21:13:45  oboehm
 * @DontProfileMe is provided now (fixes 2777814)
 *
 * Revision 1.24  2009/04/22 19:55:38  oboehm
 * only methods of classes marked with "@ProfileMe" are added
 *
 * Revision 1.23  2009/04/19 12:33:53  oboehm
 * bug with wrong range of initialization fixed
 *
 * Revision 1.22  2009/04/13 15:41:35  oboehm
 * implementation for getClasspath() now based on domains attribute
 * of ClassLoader
 *
 * Revision 1.21  2009/04/08 19:41:02  oboehm
 * addAsShutdownHook() added
 *
 * Revision 1.20  2009/04/06 07:08:53  oboehm
 * detection of never called constructors implemented
 *
 * Revision 1.19  2009/04/03 21:30:09  oboehm
 * with (Abstract)ProfileAspect it is now possible
 * to find methods/constructors which are never called
 *
 * Revision 1.18  2009/02/26 10:02:21  oboehm
 * improved JAMon detection (only JAMon 2.4 will be used if found in classpath)
 * output of ProfileStatistic improved (Label contains no longer return value;
 * this is better for sorting in Excel or OpenOffice)
 *
 * Revision 1.17  2009/02/25 21:22:18  oboehm
 * because of problems with JAMon 2.0 only JAMon 2.4 will be supported
 * (if found in the classpath)
 *
 * Revision 1.16  2008/12/28 13:31:16  oboehm
 * findbugs proofed and tests slightly improved
 *
 * Revision 1.15  2008/12/27 21:49:07  oboehm
 * ProfileStatistic can now be added as shutdown hook
 *
 * Revision 1.14  2008/12/27 21:30:49  oboehm
 * Unit added to statistic tabular table
 *
 * Revision 1.13  2008/12/27 21:16:05  oboehm
 * getStatistics() added
 *
 * Revision 1.12  2008/12/27 20:34:53  oboehm
 * statistic is now sorted after total time (in descending order)
 *
 * $Source: /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java/patterntesting/runtime/monitor/ProfileStatistic.java,v $
 */
