/**
 * $Id: SimpleProfileMonitor.java,v 1.16 2009/04/09 18:30:47 oboehm Exp $
 *
 * Copyright (c) 2008 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * (c)reated 19.12.2008 by oliver (ob@oasd.de)
 */
package patterntesting.runtime.monitor;

import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.impl.LogFactoryImpl;
import org.aspectj.lang.Signature;

import patterntesting.runtime.util.SignatureHelper;


/**
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 19.12.2008
 * @version $Revision: 1.16 $
 */
public final class SimpleProfileMonitor extends AbstractProfileMonitor {

    private static final Log log = LogFactoryImpl.getLog(SimpleProfileMonitor.class);
    private SimpleProfileMonitor parent;
    private Collection<SimpleProfileMonitor> childs;
    private final Signature label;
    /** start time in nanoseconds */
    private long startTime;
    /** the measured time in milliseconds */
    private double total;
    private double lastValue;
    private double min;
    private double max;
    /** number of calls (or "hits") */
    private int hits;
    
    public SimpleProfileMonitor() {
        this.reset();
        this.label = null;
    }
    
    public void reset() {
        this.total = 0.0;
        this.lastValue = 0.0;
        this.min = Double.MAX_VALUE;
        this.max = 0.0;
        this.hits = 0;
        this.parent = null;
        this.childs = new ConcurrentLinkedQueue<SimpleProfileMonitor>();
    }
    
    public SimpleProfileMonitor(Signature label, SimpleProfileMonitor parent) {
        this.reset();
        this.label = label;
        this.parent = parent;
        this.parent.addChild(this);
    }
    
    protected void addChild(SimpleProfileMonitor child) {
        this.childs.add(child);
    }
    
    public void addChildren(List<String> labels) {
        for (String label : labels) {
            addChild(label);
        }
    }
    
    public void addChild(String label) {
        try {
            Signature sig = SignatureHelper.getAsSignature(label);
            SimpleProfileMonitor child = new SimpleProfileMonitor(sig, this);
            addChild(child);
        } catch (Exception e) {
            log.debug("can't add " + label, e);
        }
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#getMonitors()
     */
    public ProfileMonitor[] getMonitors() {
        return (ProfileMonitor[]) childs.toArray(new SimpleProfileMonitor[childs
                .size()]);
    }
    
    /**
     * We use now the signature without the return value. That's enough.
     * 
     * @param sig
     * @return
     */
    public SimpleProfileMonitor getMonitor(Signature sig) {
//        String label = sig.toString();
//        for (SimpleProfileMonitor mon : childs) {
//            if (label.equals(mon.label.toString())) {
//                return mon;
//            }
//        }
        String label = SignatureHelper.getAsString(sig);
        for (SimpleProfileMonitor mon : childs) {
            String monLabel = SignatureHelper.getAsString(mon.label);
            if (label.equals(monLabel)) {
                return mon;
            }
        }
        return null;
    }

    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#start()
     */
    public void start() {
        this.startTime = System.nanoTime();
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#stop()
     */
    public void stop() {
        long time = System.nanoTime() - startTime;
        this.add(time / 1000000.0);
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#add(double)
     */
    public void add(double value) {
        this.lastValue = value;
        this.total += value;
        this.hits++;
        if (this.parent != null) {
            this.parent.add(value);
        }
        if (value < this.min) {
            this.min = value;
        }
        if (value > this.max) {
            this.max = value;
        }
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#getTotal()
     */
    public double getTotal() {
        return this.total;
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#getLastValue()
     */
    public double getLastValue() {
        return this.lastValue;
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#getMax()
     */
    public double getMax() {
        return this.max;
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#getMin()
     */
    public double getMin() {
        return this.min;
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#getHits()
     */
    public int getHits() {
        return this.hits;
    }
    
    /* (non-Javadoc)
     * @see patterntesting.runtime.monitor.ProfileMonitor#getAvg()
     */
    public double getAvg() {
        return this.total / this.hits;
    }
    
    public String getLabel() {
        if (this.label == null) {
            return "root";
        } else {
            return SignatureHelper.getAsString(label);
        }
    }

    @Override
    public String toString() {
        return this.getLabel() + " - " + toShortString();
    }
    
    public String toShortString() {
        return "total: " + this.total + " ms / avg: "
        + this.getAvg() + " ms / hits: " + this.hits;
    }
    
    public String toCsvHeadline() {
        return "Label; Unit; Total; Avg; Hits; Max; Min";
    }

    public String toCsvString() {
		return this.getLabel() + "; ms; " + this.total + "; "
				+ this.getAvg() + "; " + this.hits + "; " + this.getMax() + ";"
				+ this.getMin();
	}
    
    public String getUnits() {
        return "ms";
    }

    public double getActive() {
        return 0.0;
    }

    public double getAvgActive() {
        return 0.0;
    }

    public Date getFirstAccess() {
        return null;
    }

    public Date getLastAccess() {
        return null;
    }

    public double getMaxActive() {
        return 0.0;
    }
    
}

/**
 * $Log: SimpleProfileMonitor.java,v $
 * Revision 1.16  2009/04/09 18:30:47  oboehm
 * no longer double entries (bug 2745302 fixed)
 *
 * Revision 1.15  2009/04/03 21:30:09  oboehm
 * with (Abstract)ProfileAspect it is now possible
 * to find methods/constructors which are never called
 *
 * Revision 1.14  2009/03/30 06:59:45  oboehm
 * 2719516 fixed
 *
 * Revision 1.13  2009/03/07 21:03:50  oboehm
 * getShortLabel() replaced by getLabel()
 *
 * Revision 1.12  2009/02/27 20:29:07  oboehm
 * min column is also printed to CSV file
 *
 * Revision 1.11  2009/02/27 20:16:00  oboehm
 * measuring now the time in nanoseconds precision
 *
 * Revision 1.10  2009/02/26 10:02:21  oboehm
 * improved JAMon detection (only JAMon 2.4 will be used if found in classpath)
 * output of ProfileStatistic improved (Label contains no longer return value;
 * this is better for sorting in Excel or OpenOffice)
 *
 * Revision 1.9  2008/12/27 21:30:49  oboehm
 * Unit added to statistic tabular table
 *
 * Revision 1.8  2008/12/27 21:16:05  oboehm
 * getStatistics() added
 *
 * Revision 1.7  2008/12/27 20:34:53  oboehm
 * statistic is now sorted after total time (in descending order)
 *
 * Revision 1.6  2008/12/27 20:03:41  oboehm
 * Unit in dumped statistic is now a single column
 *
 * $Source: /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java/patterntesting/runtime/monitor/SimpleProfileMonitor.java,v $
 */
