/**
 * $Id: SignatureHelper.java,v 1.3 2009/04/09 18:30:47 oboehm Exp $
 *
 * Copyright (c) 2009 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * (c)reated 03.04.2009 by oliver (ob@aosd.de)
 */
package patterntesting.runtime.util;

import java.lang.reflect.*;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.impl.LogFactoryImpl;
import org.aspectj.lang.Signature;
import org.aspectj.lang.reflect.*;

import patterntesting.runtime.util.reflect.*;

/**
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 03.04.2009
 * @version $Revision: 1.3 $
 */
public class SignatureHelper {
    
    private static final Log log = LogFactoryImpl.getLog(SignatureHelper.class);

    /**
     * The difference to Signature.toString() is that in case of a methode
     * or constructor the return type is not part of result.
     * 
     * @param sig
     * @return
     */
    public static String getAsString(Signature sig) {
        try {
            CodeSignature codeSig = (CodeSignature) sig;
            String name = getAsString(sig.getDeclaringTypeName(), sig);
            Class<?>[] params = codeSig.getParameterTypes();
            return name + "(" + Converter.toShortString(params) + ")";
        } catch (Exception e) {
            log.debug("can't get '" + sig + "' as String", e);
            return sig.toString();
        }
    }
    
    public static String getAsString(String type, Signature sig) {
        String name = sig.getName();
        if (name.equals("<init>")) {
            return "new " + type;
        } else {
            return type + "." + name;
        }
    }

    public static MethodSignature getAsSignature(Method method) {
        return new MethodSignatureImpl(method);
    }
    
    public static ConstructorSignature getAsSignature(Constructor<?> ctor) {
        return new ConstructorSignatureImpl(ctor);
    }
    
    /**
     * @param label e.g. "java.lang.String.substring(int)" or
     *                   "new java.lang.String()
     * @return
     * @throws ClassNotFoundException 
     * @throws NoSuchMethodException 
     * @throws SecurityException 
     */
    public static Signature getAsSignature(String label)
            throws ClassNotFoundException, SecurityException,
            NoSuchMethodException {
        if (label.startsWith("new ")) {
            return getAsConstructorSignature(label);
        }
        String classDotMethod = StringUtils.substringBefore(label, "(");
        String className = StringUtils.substringBeforeLast(classDotMethod, ".");
        String methodName = StringUtils.substringAfterLast(classDotMethod, ".");
        Class<?>[] params = getParameterTypes(label);
        return getAsSignature(Class.forName(className), methodName, params);
    }
    
    private static Signature getAsSignature(Class<?> clazz, String methodName,
            Class<?>[] parameterTypes) throws ClassNotFoundException, SecurityException,
            NoSuchMethodException {
        Method method = clazz.getMethod(methodName, parameterTypes);
        return getAsSignature(method);
    }
    
    private static ConstructorSignature getAsConstructorSignature(String label)
            throws ClassNotFoundException, SecurityException,
            NoSuchMethodException {
        String ctor = StringUtils.substringAfterLast(label, " ").trim();
        String className = StringUtils.substringBefore(ctor, "(");
        Class<?>[] params = getParameterTypes(label);
        return getAsConstructorSignature(Class.forName(className), params);
    }
    
    private static ConstructorSignature getAsConstructorSignature(Class<?> clazz,
            Class<?>[] parameterTypes) throws ClassNotFoundException, SecurityException,
            NoSuchMethodException {
        Constructor<?> ctor = clazz.getConstructor(parameterTypes);
        return getAsSignature(ctor);
    }
    
    private static Class<?>[] getParameterTypes(String label)
            throws ClassNotFoundException {
        String parameters = StringUtils.substringAfter(label, "(");
        parameters = StringUtils.substringBefore(parameters, ")");
        String[] params = StringUtils.split(parameters, ',');
        Class<?>[] parameterTypes = new Class<?>[params.length];
        for (int i = 0; i < params.length; i++) {
            parameterTypes[i] = Class.forName(params[i].trim());
        }
        return parameterTypes;
    }

}

/**
 * $Log: SignatureHelper.java,v $
 * Revision 1.3  2009/04/09 18:30:47  oboehm
 * no longer double entries (bug 2745302 fixed)
 *
 * Revision 1.2  2009/04/06 07:08:53  oboehm
 * detection of never called constructors implemented
 *
 * Revision 1.1  2009/04/03 21:30:09  oboehm
 * with (Abstract)ProfileAspect it is now possible
 * to find methods/constructors which are never called
 *
 * $Source: /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java/patterntesting/runtime/util/SignatureHelper.java,v $
 */
